#ifdef TEST
#include <stdio.h>
#include <sys/types.h>
#include <stdint.h>
#include <memory.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/syscall.h>
#else
int syscall(int number, ...);
#include <stdlib.h>
#include <malloc.h>
#include <stdint.h>
#include <strlen.h>
#include <itox.h>
#endif

#include <mixr_structures.h>

#define PROT_NONE 0x0
#define PROT_READ 0x1
#define PROT_WRITE 0x2
#define PROT_EXEC 0x4

static void *bottom_of_stack_ptr;
static void *global_table_addr;
static int q_fd;
#ifndef TEST
#define SYS_mprotect 10
#endif

#define DT_ENTRY_SIZE (8+2+2+1)
#define DT_LENGTH_SIZE 8
#define DT_FIXED_SIZE_SIZE 8
#define DT_FIRST_ENTRY_OFFSET (DT_LENGTH_SIZE+DT_FIXED_SIZE_SIZE)

#define PAGE_SIZE 4096
#define PAGE_MASK (~(PAGE_SIZE-1))

#ifdef TEST
#define MIXR_FIXED_SIZE (sizeof(int))
#else
#define MIXR_FIXED_SIZE (60)
#endif

#ifdef DEBUG
	#define print_str_debug mixr_print_str
	#define print_int_debug mixr_print_int
	#define print_addr_debug mixr_print_addr
	#define print_unsigned_long_long_debug mixr_print_unsigned_long_long
#else
	#define print_str_debug(a) 
	#define print_int_debug(a) 
	#define assert(a)
#endif

#ifndef POLICY_RANDOM_MAX_ITERATIONS 
#define POLICY_RANDOM_MAX_ITERATIONS 10
#endif

int _mprotect(void *addr, size_t len, int flags);
void _memcpy(void *dst, const void *src, unsigned long int len);

uint8_t swap_space[MIXR_FIXED_SIZE];

uint64_t dt_table_length(void *);
uint64_t dt_table_read_entry_address(void *, unsigned int );
void dt_table_write_entry_active(void *, unsigned int, uint8_t );
#ifdef DEBUG
#define write_fd 2
void mixr_print_str(char *s)
{
	write(write_fd,s,strlen(s));
}
void mixr_print_int(int x)
{
	char buf[100];
	itox(x,buf);
	write(write_fd,buf,strlen(buf));
}
void mixr_print_unsigned_long_long(unsigned long long x)
{
	char buf[100];
	itox(x,buf);
	write(write_fd,buf,strlen(buf));
}
void mixr_print_addr(void *x)
{
	mixr_print_unsigned_long_long((unsigned long long)x);
}
#endif

void mixr_active_dollops_clear(void *table_addr) {
	uint64_t table_length = dt_table_length(table_addr);
	unsigned int table_iterator = 0;
	volatile uint8_t empty = 0;

	_mprotect((void*)(((uintptr_t)table_addr)&PAGE_MASK),
	          (table_length*DT_ENTRY_SIZE) +
						 DT_FIXED_SIZE_SIZE +
						 DT_LENGTH_SIZE +
						 PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);
	for (table_iterator = 0; table_iterator<table_length; table_iterator++)
	{
		dt_table_write_entry_active(table_addr, table_iterator, empty);
	}
	_mprotect((void*)(((uintptr_t)table_addr)&PAGE_MASK),
	          (table_length*DT_ENTRY_SIZE) +
						 DT_FIXED_SIZE_SIZE +
						 DT_LENGTH_SIZE +
						 PAGE_SIZE,
						PROT_READ|PROT_EXEC);
}

unsigned int mixr_find_dollop_containing(void *table_addr, uint64_t addr, uint8_t *found)
{
	uint64_t table_length = dt_table_length(table_addr);
	unsigned int table_iterator = 0;
	uint64_t table_iterator_address = 0;

	*found = 0;

	for (table_iterator = 0; table_iterator<table_length; table_iterator++)
	{
		table_iterator_address = dt_table_read_entry_address(table_addr, table_iterator);

		if (table_iterator_address<=addr && addr<=(table_iterator_address+MIXR_FIXED_SIZE)) {
			*found = 1;
			return table_iterator;
		}
	}
	return 0;
}

void mixr_active_dollops_mark_from_stack(void *table_addr) {
	void *stack_ptr = NULL;
	/*
	 * There are other hacks to get this, but
	 * let's just use the real thing.
	 */
	asm ("movq %%rsp, %0\n"
	: "=r" (stack_ptr)
	:);

	_mprotect((void*)(((uintptr_t)table_addr)&PAGE_MASK),
	          (dt_table_length(table_addr)*DT_ENTRY_SIZE) +
						 DT_FIXED_SIZE_SIZE +
						 DT_LENGTH_SIZE +
						 PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);

	while (stack_ptr != bottom_of_stack_ptr)
	{
		uint64_t stack_value = *((uint64_t*)stack_ptr);
		uint8_t found = 0;	
		unsigned int entry = 0;

		entry = mixr_find_dollop_containing(table_addr, stack_value, &found);
#ifdef DEBUG
#ifdef TEST
		printf("stack_ptr: %lx @ 0x%lx\n", stack_value, stack_ptr);
#else
		print_str_debug("stack: ");
		print_int_debug(stack_value);
		print_str_debug(" @ 0x");
		print_addr_debug(stack_ptr);
		print_str_debug("\n");
#endif
#endif
		if (found)
		{
#ifdef DEBUG
#ifdef TEST
			printf("Pointer to dollop %lu at 0x%lx\n", entry, stack_ptr);
#else
			print_str_debug("Pointer to dollop ");
			print_int_debug(entry);
			print_str_debug(" @ 0x");
			print_addr_debug(stack_ptr);
			print_str_debug("\n");
#endif
#endif
			dt_table_write_entry_active(table_addr, entry, 1);
		}
		/*
		 * In x86-64 (and perhaps earlier), stack is 
		 * 8-byte aligned. Let's walk at that 
		 * resolution.
		 */
		stack_ptr+=0x8;
	}
	_mprotect((void*)(((uintptr_t)table_addr)&PAGE_MASK),
	          (dt_table_length(table_addr)*DT_ENTRY_SIZE) +
						 DT_FIXED_SIZE_SIZE +
						 DT_LENGTH_SIZE +
						 PAGE_SIZE,
						PROT_READ|PROT_EXEC);
}

uint64_t read_random64(void) {
	unsigned long long random_value = -1;
	int fd = -1;
#ifdef CFAR
	fd = open("/dev/cfar_urandom", 0x00, 0x00);
#else
	fd = open("/dev/urandom", 0x00, 0x00);
#endif
	if (fd != -1) {
		read(fd, (void*)&random_value, sizeof(unsigned long long));
		close(fd);
	}
	return random_value;
}

#ifdef TEST
#define _memcpy memcpy
#else
void _memcpy(void *dst, const void *src, unsigned long int len) 
{
	unsigned long int i = 0;
	for (i = 0; i<len; i++) {
		((uint8_t*)(dst))[i] = ((uint8_t*)(src))[i];
	}
	return;
}
#endif

void make_writable(void *addr, size_t len)
{
}

void make_unwritable(void *addr, size_t len)
{
}

int _mprotect(void *addr, size_t len, int flags)
{
#ifdef TEST
	printf("_mprotect: addr: %p\n", addr);
	//return mprotect(addr, len, flags);
	return 0;
#else
	return syscall(SYS_mprotect, addr, len, flags);
#endif
}

uint16_t de_table_entry_count(void *de_table_addr)
{
	uint16_t entry_count = 0;

	_memcpy(&entry_count, (const void*)de_table_addr, sizeof(uint16_t));

	return entry_count;
}

uint16_t rw_table_entry_count(void *rw_table_addr)
{
	uint16_t entry_count = 0;

	_memcpy(&entry_count, (const void*)rw_table_addr, sizeof(uint16_t));

	return entry_count;
}

struct de_table_entry de_table_read_entry(void *de_table_addr, unsigned int entry)
{
	struct de_table_entry de_entry = {};
	void *entry_address = NULL;
	uint16_t entry_count = de_table_entry_count(de_table_addr);

	if (entry_count < entry)
		return de_entry;

	entry_address = (void*)(de_table_addr + 
	                        entry*DE_TABLE_ENTRY_SIZE +
	                        DE_TABLE_HEADER_SIZE);
	_memcpy(&de_entry, entry_address, sizeof(struct de_table_entry));

	return de_entry;
}

void de_table_write_entry(void *de_table_addr, unsigned int entry, struct de_table_entry de_entry)
{
	void *entry_address = NULL;
	uint16_t entry_count = de_table_entry_count(de_table_addr);

	if (entry_count < entry)
		return;

	entry_address = (void*)(de_table_addr + 
	                        entry*DE_TABLE_ENTRY_SIZE +
	                        DE_TABLE_HEADER_SIZE);
	_memcpy(entry_address, &de_entry, sizeof(struct de_table_entry));
}

struct rw_table_entry rw_table_read_entry(void *rw_table_addr, unsigned int entry)
{
	struct rw_table_entry rw_entry = {};
	void *entry_address = NULL;
	uint16_t entry_count = rw_table_entry_count(rw_table_addr);

	if (entry_count < entry)
		return rw_entry;

	entry_address = (void*)(rw_table_addr + 
	                        entry*RW_TABLE_ENTRY_SIZE +
	                        RW_TABLE_HEADER_SIZE);
	_memcpy(&rw_entry, entry_address, sizeof(struct rw_table_entry));

	return rw_entry;
}

void rw_table_write_entry(void *rw_table_addr, unsigned int entry, struct rw_table_entry rw_entry)
{
	void *entry_address = NULL;
	uint16_t entry_count = rw_table_entry_count(rw_table_addr);

	if (entry_count < entry)
		return;

	entry_address = (void*)(rw_table_addr + 
	                        entry*RW_TABLE_ENTRY_SIZE +
	                        RW_TABLE_HEADER_SIZE);
	_memcpy(entry_address, &rw_entry, sizeof(struct rw_table_entry));
}

#ifdef TEST
void rw_table_print_entry(struct rw_table_entry rw_entry)
{
	printf("RW Entry: {%hd, 0x%hhx, 0x%hhx, 0x%lx}\n", rw_entry.instruction_offset,rw_entry.rewrite_offset, rw_entry.rewrite_length,rw_entry.absolute_value);
}
void de_table_print_entry(struct de_table_entry de_entry)
{
	printf("DE Entry: {%hx, 0x%hhx, 0x%x}\n", de_entry.tag,de_entry.instruction_offset,de_entry.dollop_table_ptr);
}
#endif

uint64_t dt_table_length(void *dt_table_addr)
{
	uint64_t table_length = 0;

	_memcpy(&table_length, (const void*)dt_table_addr, sizeof(uint64_t));

	return table_length;
}
uint64_t dt_fixed_size(void *dt_table_addr)
{
	uint64_t fixed_size = 0;

	_memcpy(&fixed_size,
	       (const void*)(((uint64_t)dt_table_addr) + DT_LENGTH_SIZE),
				 sizeof(uint64_t));

	return fixed_size;
}

int dt_table_write_to_file(void *dt_table_addr, char *filename)
{
	uint64_t entry_count = dt_table_length(dt_table_addr);
	uint64_t entry_size = dt_fixed_size(dt_table_addr);
	int fd = -1;
	
	fd = open(filename, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR);
	if (fd == -1)
		return -1;
	
	write(fd, dt_table_addr, 8 /* entry count*/ + 
	                         8 /* entry size*/ + 
													 (8+2+1)*entry_count);
	close(fd);
}

int mixr_open_fifo_q(char *q_name)
{
	mkfifo(q_name, S_IRUSR|S_IWUSR);
	return open(q_name, O_RDWR, S_IRUSR|S_IWUSR);
}

int mixr_close_fifo_q(int q)
{
	return close(q);
}

uint64_t dt_table_entry_address(void *dt_table_addr, unsigned int entry)
{
	uint64_t entry_offset = entry * DT_ENTRY_SIZE + DT_FIRST_ENTRY_OFFSET; 
	return (uint64_t)(dt_table_addr+entry_offset);
}

uint64_t dt_table_read_entry_address(void *dt_table_addr, unsigned int entry)
{
	uint64_t entry_address = 0;

	_memcpy(&entry_address,
	       (const void*)dt_table_entry_address(dt_table_addr, entry),
				 sizeof(uint64_t));

	return entry_address;
}

uint16_t dt_table_read_de_entry_offset(void *dt_table_addr, unsigned int entry)
{
	uint16_t entry_offset;
	_memcpy(&entry_offset,
	        (const void*)(dt_table_entry_address(dt_table_addr, entry)+8),
					sizeof(uint16_t));
	return entry_offset;
}

uint16_t dt_table_read_rw_entry_offset(void *dt_table_addr, unsigned int entry)
{
	uint16_t entry_offset;
	_memcpy(&entry_offset,
	        (const void*)(dt_table_entry_address(dt_table_addr, entry)+8+2),
					sizeof(uint16_t));
	return entry_offset;
}

uint8_t dt_table_read_entry_active(void *dt_table_addr, unsigned int entry)
{
	uint8_t entry_active;
	_memcpy(&entry_active,
	        (const void*)(dt_table_entry_address(dt_table_addr, entry)+12),
					sizeof(uint8_t));
	return entry_active;
}
void dt_table_write_entry_address(void *dt_table_addr, unsigned int entry, uint64_t entry_address)
{
	_memcpy((void*)dt_table_entry_address(dt_table_addr, entry), 
	       (const void*)&entry_address,
				 sizeof(uint64_t));
}

void dt_table_write_de_entry_offset(void *dt_table_addr, unsigned int entry, uint16_t entry_offset)
{
	_memcpy((void*)(dt_table_entry_address(dt_table_addr, entry)+8),
	       (const void*)&entry_offset,
				 sizeof(uint16_t));
}

void dt_table_write_rw_entry_offset(void *dt_table_addr, unsigned int entry, uint16_t entry_offset)
{
	_memcpy((void*)(dt_table_entry_address(dt_table_addr, entry)+8+2),
	       (const void*)&entry_offset,
				 sizeof(uint16_t));
}

void dt_table_write_entry_active(void *dt_table_addr, unsigned int entry, uint8_t entry_active)
{

	_memcpy((void*)(dt_table_entry_address(dt_table_addr, entry)+12), 
	       (const void*)&entry_active,
				 sizeof(uint8_t));
}

void dt_table_swap_entry_addresses(void *dt_table_addr, unsigned int a, unsigned int b)
{
	uint64_t address_entry_a = dt_table_read_entry_address(dt_table_addr, a);
	uint64_t address_entry_b = dt_table_read_entry_address(dt_table_addr, b);

	dt_table_write_entry_address(dt_table_addr, a, address_entry_b);
	dt_table_write_entry_address(dt_table_addr, b, address_entry_a);
}

void dt_table_swap_entry_contents(void *dt_table_addr, unsigned int a, unsigned int b)
{
	uint64_t fixed_size = MIXR_FIXED_SIZE;
	//void *swap_space = (void*)malloc(fixed_size);
	//uint8_t swap_space[MIXR_FIXED_SIZE] = {0,};
	void *a_contents_addr = (void*)dt_table_read_entry_address(dt_table_addr,a);
	void *b_contents_addr = (void*)dt_table_read_entry_address(dt_table_addr,b);
	uint64_t table_length = (dt_table_length(dt_table_addr)*DT_ENTRY_SIZE) +
	                         DT_FIXED_SIZE_SIZE +
													 DT_LENGTH_SIZE;
	/*
	 * Three steps:
	 * 1. mprotect + w
	 * 2. copy
	 * 3. mprotect - w
	 * The assumption is that we are '-w' in these
	 * pages when we start.
	 */
	_mprotect((void*)(((uintptr_t)dt_table_addr)&PAGE_MASK),
	          table_length + PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);
	_mprotect((void*)(((uintptr_t)a_contents_addr)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);
	_mprotect((void*)(((uintptr_t)b_contents_addr)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);
#ifdef DEBUG
	print_str_debug("swap_space: ");
	print_addr_debug((void*)(&swap_space));
	print_str_debug("\n");
	print_str_debug("a_contents_addr: ");
	print_addr_debug(a_contents_addr);
	print_str_debug("\n");
	print_str_debug("b_contents_addr: ");
	print_addr_debug(b_contents_addr);
	print_str_debug("\n");
	print_str_debug("fixed_size: ");
	print_int_debug(fixed_size);
	print_str_debug("\n");
#endif

	_memcpy(&swap_space, a_contents_addr, fixed_size);
	_memcpy(a_contents_addr, b_contents_addr, fixed_size); 
	_memcpy(b_contents_addr, &swap_space, fixed_size); 

#ifdef DEBUG
	print_str_debug("a: ");
	print_int_debug(dt_table_read_entry_address(dt_table_addr, a));
	print_str_debug("\n");
	print_str_debug("b: ");
	print_int_debug(dt_table_read_entry_address(dt_table_addr, b));
	print_str_debug("\n");
#endif

	dt_table_swap_entry_addresses(dt_table_addr, a, b);

#ifdef DEBUG
	print_str_debug("a: ");
	print_int_debug(dt_table_read_entry_address(dt_table_addr, a));
	print_str_debug("\n");
	print_str_debug("b: ");
	print_int_debug(dt_table_read_entry_address(dt_table_addr, b));
	print_str_debug("\n");
#endif

	_mprotect((void*)((uintptr_t)(dt_table_addr)&PAGE_MASK),
	          table_length + PAGE_SIZE,
						PROT_READ|PROT_EXEC);
	_mprotect((void*)((uintptr_t)(a_contents_addr)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC);
	_mprotect((void*)((uintptr_t)(b_contents_addr)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC);
}

void mixr_do_de(uint64_t dollop_start, struct de_table_entry de_entry)
{
	uint64_t next_pc_address = dollop_start + de_entry.instruction_offset;
	uint64_t dest = next_pc_address - 0x4;
	uint32_t jump_target = 0;

	_memcpy((void*)&jump_target, (uint32_t*)de_entry.dollop_table_ptr, sizeof(uint32_t));

#if DEBUG
	print_str_debug("DE Table Entry through ");
	print_unsigned_long_long_debug(de_entry.dollop_table_ptr);
	print_str_debug(" to (0x");
	print_addr_debug((void*)jump_target);
	print_str_debug(") written to ");
	print_addr_debug((void*)dest);
	print_str_debug("\n");
#endif

	jump_target -= next_pc_address;

#ifdef DEBUG
	if (jump_target == NULL)
		jump_target = 0xdead;
#else
	assert(jump_target);
#endif

	_mprotect((void*)(((uintptr_t)dest)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);
	_memcpy((void*)dest, &jump_target, sizeof(uint32_t));
	_mprotect((void*)(((uintptr_t)dest)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC);
}

void mixr_do_rw(uint64_t dollop_start, struct rw_table_entry rw_entry)
{
	uint64_t next_pc_address = dollop_start + rw_entry.instruction_offset;
	uint64_t relative = next_pc_address - rw_entry.rewrite_offset;
	void *rw_val = NULL;
	uint8_t rw8_val;
	uint16_t rw16_val;
	uint32_t rw32_val;
	uint64_t rw48_val;
	uint64_t rw64_val;
	int rewrite_length = 0;

#if DEBUG
	print_str_debug("Abs ");
	print_unsigned_long_long_debug(rw_entry.absolute_value);
	print_str_debug(" to (0x");
	print_addr_debug((void*)next_pc_address);
	print_str_debug(") ");
	print_int_debug(rw_entry.absolute_value-next_pc_address);
	print_str_debug(" at ");
	print_addr_debug((void*)relative);
	print_str_debug("\n");
#endif

	switch (rw_entry.rewrite_length)
	{
		case 0x0:
		{
			rw8_val = rw_entry.absolute_value-next_pc_address;
			rw_val = (void*)&rw8_val;
			rewrite_length = 1;
			break;
		}
		case 0x1:
		{
			rw16_val = rw_entry.absolute_value-next_pc_address;
			rw_val = (void*)&rw16_val;
			rewrite_length = 2;
			break;
		}
		case 0x3:
		{
			rw32_val = rw_entry.absolute_value-next_pc_address;
			rw_val = (void*)&rw32_val;
			rewrite_length = 4;
			break;
		}
		case 0x5:
		{
			rw48_val = rw_entry.absolute_value-next_pc_address;
			rw_val = (void*)&rw48_val;
			rewrite_length = 6;
			break;
		}
		case 0x7:
		{
			rw64_val = rw_entry.absolute_value-next_pc_address;
			rw_val = (void*)&rw64_val;
			rewrite_length = 8;
			break;
		}
	}
#ifdef DEBUG
	if (rw_val == NULL)
		rw_val = 0xdead;
#else
	assert(rw_val);
#endif

	_mprotect((void*)(((uintptr_t)relative)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC|PROT_WRITE);
	_memcpy((void*)relative, rw_val, rewrite_length);
	_mprotect((void*)(((uintptr_t)relative)&PAGE_MASK),
	          PAGE_SIZE + PAGE_SIZE,
						PROT_READ|PROT_EXEC);
}

#define get_rbp(output) \
{ \
	uint64_t _rbp_value = 0xfeed; \
	asm volatile ("mov %%rbp, %0\n" \
	    : "=r" (_rbp_value) \
	   ); \
	output = _rbp_value; \
} \

#define get_rsp(output) \
{ \
	uint64_t _rsp_value = 0xfeed; \
	asm volatile ("mov %%rsp, %0\n" \
			: "=r" (_rsp_value) \
			); \
	output = _rsp_value; \
} \



void zipr_mixr(void *table_addr)
{
	uint64_t table_length = 0;
#ifdef POLICY_RANDOM
	uint64_t random_iterator = 0;
	uint64_t random_iterator_max = POLICY_RANDOM_MAX_ITERATIONS;
#else
	uint64_t table_iterator = 2;
#endif
	uint64_t table_fixed_size = 0;

	uint64_t q_value = 0xab;
	
	table_length = dt_table_length(table_addr);
	table_fixed_size = dt_fixed_size(table_addr);

#ifdef DEBUG
		print_str_debug("dt_fixed_size:");
		print_unsigned_long_long_debug(table_fixed_size);
		print_str_debug("\n");
#endif

	mixr_active_dollops_mark_from_stack(table_addr);

#ifdef POLICY_RANDOM
	for (random_iterator=0;random_iterator<random_iterator_max;random_iterator++)
#else
	for (table_iterator=1;table_iterator<table_length;table_iterator+=2)
#endif
	{
#ifdef POLICY_RANDOM
		uint64_t a = read_random64() % table_length;
		uint64_t b = read_random64() % table_length;
#else
		uint64_t a = table_iterator, b = a-1;
#endif

		/*
		 * Calculations of addresses/pointers for dollop A
		 */
		uint64_t dt_address_a = dt_table_read_entry_address(table_addr, a);
		uint16_t dt_rw_offset_a = dt_table_read_rw_entry_offset(table_addr, a);
		uint64_t rw_address_a = dt_address_a + dt_rw_offset_a;
		uint16_t rw_entry_count_a = rw_table_entry_count((void*)rw_address_a);

		/*
		 * Calculations of addresses/pointers for dollop B
		 */
		uint64_t dt_address_b = dt_table_read_entry_address(table_addr, b);
		uint16_t dt_rw_offset_b = dt_table_read_rw_entry_offset(table_addr, b);
		uint64_t rw_address_b = dt_address_b + dt_rw_offset_b;
		uint16_t rw_entry_count_b = rw_table_entry_count((void*)rw_address_b);

		uint16_t rw_entry_count_iterator = 0;

#ifdef DEBUG
		print_str_debug("Swapping:");
		print_unsigned_long_long_debug(a);
		print_str_debug(": (");
		print_addr_debug((void*)dt_address_a);
		print_str_debug(",");
		print_int_debug(dt_rw_offset_a);
		print_str_debug(" #");
		print_int_debug(rw_entry_count_a);
		print_str_debug(") <-> ");
		print_int_debug(b);
		print_str_debug(": (");
		print_addr_debug((void*)dt_address_b);
		print_str_debug(",");
		print_int_debug(dt_rw_offset_b);
		print_str_debug(" #");
		print_int_debug(rw_entry_count_b);
		print_str_debug(")\n");
#endif
		if (dt_table_read_entry_active(table_addr, a) ||
		    dt_table_read_entry_active(table_addr, b))
		{
#ifdef DEBUG
			print_str_debug("Skipping swap ");
			print_int_debug(a);
			print_str_debug("<->");
			print_int_debug(b);
			print_str_debug(" because it is marked active.\n");
#endif
			continue;
		}

		if (a == b) 
		{
#ifdef DEBUG
			print_str_debug("Skipping swap from ourself to ourself\n");
#endif
			continue;
		}

		dt_table_swap_entry_contents(table_addr, a, b);

		/*
		 * Rewrite a.
		 */
		for (rw_entry_count_iterator = 0;
		     rw_entry_count_iterator<rw_entry_count_a;
				 rw_entry_count_iterator++)
		{
			/*
			 * Since we already swapped, the old a is now
			 * at b. To reuse variable assignments from above,
			 * this looks more confusing.
			 */
			mixr_do_rw(dt_address_b,
			           rw_table_read_entry((void*)(dt_address_b+dt_rw_offset_a),
								                     rw_entry_count_iterator));
		}

		/*
		 * Rewrite b.
		 */
		for (rw_entry_count_iterator = 0;
		     rw_entry_count_iterator<rw_entry_count_b;
				 rw_entry_count_iterator++)
		{
			/*
			 * Since we already swapped, the old b is now
			 * at a. To reuse variable assignments from above,
			 * this looks more confusing.
			 */
			mixr_do_rw(dt_address_a,
			           rw_table_read_entry((void*)(dt_address_a+dt_rw_offset_b),
								                     rw_entry_count_iterator));
		}
	}
#ifdef DEBUG
	print_str_debug("Done with the swapping!\n");
#endif
	for (table_iterator=0;table_iterator<table_length;table_iterator++)
	{
		uint64_t dt_address=dt_table_read_entry_address(table_addr, table_iterator);
		uint16_t dt_de_offset = dt_table_read_de_entry_offset(table_addr,
		                                                      table_iterator);
		uint64_t de_address = dt_address + dt_de_offset;
		uint16_t de_entry_count = de_table_entry_count((void*)de_address);
		uint16_t de_entry_count_iterator = 0;

		for (de_entry_count_iterator = 0;
		     de_entry_count_iterator<de_entry_count;
				 de_entry_count_iterator++)
		{
			struct de_table_entry entry = de_table_read_entry((void*)de_address,
								                    de_entry_count_iterator);
			if (entry.tag != 0x0)
			{
#ifdef DEBUG
				print_str_debug("Skipping.\n");
#endif
				continue;
			}	

#ifdef DEBUG
			print_str_debug("Looking at ");
			print_addr_debug(de_address);
			print_str_debug("'s ");
			print_int_debug(de_entry_count_iterator);
			print_str_debug(" entry.\n");
#endif
			if (entry.dollop_table_ptr == 0x0)
			{
#ifdef DEBUG
				print_str_debug("Skipping because it is empty!\n");
#endif
				continue;
			}
			mixr_do_de(dt_address, entry);
		}
	}

#ifdef DEBUG
	print_str_debug("Done with the fixing up DE Table entries!\n");
#endif

	mixr_active_dollops_clear(table_addr);
#ifdef DEBUG
	print_str_debug("Done with the clearing!\n");
#endif

#ifdef DEBUG
	print_str_debug("Done with swapping!\n");
#endif

#ifdef DEMO
	dt_table_write_to_file(table_addr, "/tmp/testwrite.bin");

	if (q_fd == -1)
	{
		print_str_debug("Attempting to open fifo q file handle.");
		q_fd = mixr_open_fifo_q("/tmp/testfifo.txt");
	}
	print_str_debug("Attempting to write to fifo q file handle:");
	print_int_debug(q_fd);
	print_str_debug("\n");
	write(q_fd, (const char *)&q_value, sizeof(uint65_t));	
#endif
}

void zipr_hook_start(unsigned long long id,
                     unsigned long long rax,
										 unsigned long long rsp)
{
	global_table_addr = 0;
	uint64_t ra = 0, rbp_value = 0, rsp_value = 0;

	//a = read_random64() % table_length;
	//b = read_random64() % table_length;

	q_fd = -1;

	get_rbp(rbp_value);

	get_rsp(rsp_value);
	bottom_of_stack_ptr = (void*)rbp_value;

	/*
	 * The return value is 16 bytes above
	 * the rbp.
	 */
	ra = *((uint64_t*)(rbp_value+0x8));

#ifdef DEBUG
	print_str_debug("rbp_value:");
	print_unsigned_long_long_debug(rbp_value);
	print_str_debug("\n");
	print_str_debug("ra:");
	print_unsigned_long_long_debug(ra);
	print_str_debug("\n");
	print_str_debug("global_table_addr:");
	print_unsigned_long_long_debug(id);
	print_str_debug("\n");
#endif

	global_table_addr = (void*)id;
	zipr_mixr(global_table_addr);
}

void zipr_hook_dynamic_callback(unsigned int id, unsigned long long rax, unsigned long long rsp)
{
	print_str_debug("zipr_hook_dynamic_callback()\n");
	//zipr_mixr(global_table_addr);
}

#ifdef TEST
void test_dollop_table(void) 
{
	uint64_t table_entries = 11;
	uint64_t table_fixed_size = MIXR_FIXED_SIZE;
	uint64_t table_entry_address = 0xff00ff00;
	uint16_t table_entry_offset = 0;
	void *table = (void*)malloc((table_entries*DT_ENTRY_SIZE)+DT_FIRST_ENTRY_OFFSET);
	unsigned i = 0;

	int nine = 9;
	int ten = 10;

	uint16_t nine_de_offset = 0x99;
	uint16_t ten_de_offset = 0xaa;

	uint16_t nine_rw_offset = 0x9;
	uint16_t ten_rw_offset = 0xa;

	uint8_t nine_active = 0x9;
	uint8_t ten_active = 0xa;

	_memcpy(table, (const void*)&table_entries, sizeof(uint64_t));
	_memcpy(table, (const void*)&table_entries, sizeof(uint64_t));
	_memcpy((void*)(((uint64_t)table)+(DT_LENGTH_SIZE)),
	       (const void*)&table_fixed_size,
				 sizeof(uint64_t));

	for (i = 0; i<table_entries-2; i++)
	{
		table_entry_address += 1;
		table_entry_offset += 1;
		printf("table_entry_address: (%d): %lx\n", i, table_entry_address);
		printf("[de|rw]_table_entry_offset: (%d): %hx\n", i, table_entry_offset);
		dt_table_write_entry_address(table, i, table_entry_address);
		dt_table_write_de_entry_offset(table, i, table_entry_offset);
		dt_table_write_rw_entry_offset(table, i, table_entry_offset);
	}

	dt_table_write_entry_address(table, 9, (uint64_t)&nine);
	dt_table_write_de_entry_offset(table, 9, nine_de_offset);
	dt_table_write_rw_entry_offset(table, 9, nine_rw_offset);
	dt_table_write_entry_active(table, 9, nine_active);

	dt_table_write_entry_address(table, 10, (uint64_t)&ten);
	dt_table_write_de_entry_offset(table, 10, ten_de_offset);
	dt_table_write_rw_entry_offset(table, 10, ten_rw_offset);
	dt_table_write_entry_active(table, 10, ten_active);

	printf("dt_table_length: %lu\n", dt_table_length(table));
	printf("dt_fixed_size: %lu\n", dt_fixed_size(table));
	printf("dt_table_read_entry_address(0): %lx\n", dt_table_read_entry_address(table, 0));
	printf("dt_table_read_entry_address(5): %lx\n", dt_table_read_entry_address(table, 5));
	printf("dt_table_read_entry_address(8): %lx\n", dt_table_read_entry_address(table, 8));
	printf("dt_table_read_de_entry_offset(0): %hx\n", dt_table_read_de_entry_offset(table, 0));
	printf("dt_table_read_de_entry_offset(5): %hx\n", dt_table_read_de_entry_offset(table, 5));
	printf("dt_table_read_de_entry_offset(8): %hx\n", dt_table_read_de_entry_offset(table, 8));
	printf("dt_table_read_rw_entry_offset(0): %hx\n", dt_table_read_rw_entry_offset(table, 0));
	printf("dt_table_read_rw_entry_offset(5): %hx\n", dt_table_read_rw_entry_offset(table, 5));
	printf("dt_table_read_rw_entry_offset(8): %hx\n", dt_table_read_rw_entry_offset(table, 8));
	printf("dt_table_read_entry_active(0): %hx\n", dt_table_read_entry_active(table, 0));
	printf("dt_table_read_entry_active(5): %hx\n", dt_table_read_entry_active(table, 5));
	printf("dt_table_read_entry_active(8): %hx\n", dt_table_read_entry_active(table, 8));

	printf("swap 5 <-> 8\n");
	dt_table_swap_entry_addresses(table, 5, 8);
	
	printf("dt_table_read_entry_address(5): %lx\n", dt_table_read_entry_address(table, 5));
	printf("dt_table_read_de_entry_offset(5): %hx\n", dt_table_read_de_entry_offset(table, 5));
	printf("dt_table_read_rw_entry_offset(5): %hx\n", dt_table_read_rw_entry_offset(table, 5));
	printf("dt_table_read_entry_address(8): %lx\n", dt_table_read_entry_address(table, 8));
	printf("dt_table_read_de_entry_offset(8): %hx\n", dt_table_read_de_entry_offset(table, 8));
	printf("dt_table_read_rw_entry_offset(8): %hx\n", dt_table_read_rw_entry_offset(table, 8));

	printf("swap 8 <-> 0\n");
	dt_table_swap_entry_addresses(table, 0, 8);
	printf("dt_table_read_entry_address(0): %lx\n", dt_table_read_entry_address(table, 0));
	printf("dt_table_read_de_entry_offset(0): %hx\n", dt_table_read_de_entry_offset(table, 0));
	printf("dt_table_read_rw_entry_offset(0): %hx\n", dt_table_read_rw_entry_offset(table, 0));
	printf("dt_table_read_entry_address(8): %lx\n", dt_table_read_entry_address(table, 8));
	printf("dt_table_read_de_entry_offset(8): %hx\n", dt_table_read_de_entry_offset(table, 8));
	printf("dt_table_read_rw_entry_offset(8): %hx\n", dt_table_read_rw_entry_offset(table, 8));
	
	printf("dt_table_read_entry_address(9): %lx, %x\n", dt_table_read_entry_address(table, 9), *((int*)dt_table_read_entry_address(table, 9)));
	printf("dt_table_read_de_entry_offset(9): %hx\n", dt_table_read_de_entry_offset(table, 9));
	printf("dt_table_read_rw_entry_offset(9): %hx\n", dt_table_read_rw_entry_offset(table, 9));
	printf("dt_table_read_entry_active(9): %hx\n", dt_table_read_entry_active(table, 9));
	printf("dt_table_read_entry_address(10): %lx, %x\n", dt_table_read_entry_address(table, 10), *((int*)dt_table_read_entry_address(table, 10)));
	printf("dt_table_read_de_entry_offset(10): %hx\n", dt_table_read_de_entry_offset(table, 10));
	printf("dt_table_read_rw_entry_offset(10): %hx\n", dt_table_read_rw_entry_offset(table, 10));
	printf("dt_table_read_entry_active(10): %hx\n", dt_table_read_entry_active(table, 10));

	dt_table_swap_entry_contents(table, 10, 9);

	printf("dt_table_read_entry_address(9): %lx, %x\n", dt_table_read_entry_address(table, 9), *((int*)dt_table_read_entry_address(table, 9)));
	printf("dt_table_read_de_entry_offset(9): %hx\n", dt_table_read_de_entry_offset(table, 9));
	printf("dt_table_read_rw_entry_offset(9): %hx\n", dt_table_read_rw_entry_offset(table, 9));
	printf("dt_table_read_entry_active(9): %hx\n", dt_table_read_entry_active(table, 9));
	printf("dt_table_read_entry_address(10): %lx, %x\n", dt_table_read_entry_address(table, 10), *((int*)dt_table_read_entry_address(table, 10)));
	printf("dt_table_read_de_entry_offset(10): %hx\n", dt_table_read_de_entry_offset(table, 10));
	printf("dt_table_read_rw_entry_offset(10): %hx\n", dt_table_read_rw_entry_offset(table, 10));
	printf("dt_table_read_entry_active(10): %hx\n", dt_table_read_entry_active(table, 10));

	dt_table_write_to_file(table, "/tmp/testtable.bin");
}

void test_de_table(void)
{
	struct de_table_header *de_table = NULL;

	int entry_count = 11;
	struct de_table_entry entry_ten = {1, 2, 4};
	struct de_table_entry entry_eleven = {11, 22, 33};
	struct de_table_entry entry_fiftyk = {111, 222, 333};
	
	de_table = (struct de_table_header*)malloc(sizeof(struct de_table_header) + (entry_count*sizeof(struct de_table_entry)));

	de_table->entry_count = 11;

	de_table_write_entry((void*)de_table, 10, entry_ten);
	de_table_write_entry((void*)de_table, 11, entry_eleven);
	
	de_table_write_entry((void*)de_table, 50000, entry_fiftyk);

	de_table_print_entry(de_table_read_entry((void*)de_table, 10));
	de_table_print_entry(de_table_read_entry((void*)de_table, 11));

	/*
	 * Will return 0'd de_table_entry -- entry too high.
	 */
	de_table_print_entry(de_table_read_entry((void*)de_table, 50000));

	printf("de_table_entry_count: %d\n", de_table_entry_count((void*)de_table));
}

void test_rw_table(void)
{
	struct rw_table_header *rw_table = NULL;
	int entry_count = 11;
	struct rw_table_entry entry_ten = {1, 2, 3, 4};
	struct rw_table_entry entry_eleven = {11, 22, 33, 44};
	struct rw_table_entry entry_fiftyk = {111, 222, 333, 444};

	rw_table = (struct rw_table_header*)malloc(sizeof(struct rw_table_header) + (entry_count*sizeof(struct rw_table_entry)));

	rw_table->entry_count = 11;

	rw_table_write_entry((void*)rw_table, 10, entry_ten);
	rw_table_write_entry((void*)rw_table, 11, entry_eleven);

	/*
	 * Will fail out -- entry too high.
	 */
	rw_table_write_entry((void*)rw_table, 50000, entry_fiftyk);

	rw_table_print_entry(rw_table_read_entry((void*)rw_table, 10));
	rw_table_print_entry(rw_table_read_entry((void*)rw_table, 11));

	/*
	 * Will return 0'd rw_table_entry -- entry too high.
	 */
	rw_table_print_entry(rw_table_read_entry((void*)rw_table, 50000));
}

int main()
{
	test_de_table();
	//test_rw_table();
	//test_dollop_table();
	return 0;
}
#endif
